/*
** Portable ChaCha20/8 implementation. Based on RFC 7539.
** Copyright (C) 2015-2016 Austin Seipp. Released in the Public Domain.
*/

#include "chacha20.h"

static unsigned int
l32(unsigned int x,unsigned int n)
{
  /* See: http://blog.regehr.org/archives/1063 */
  return (x<<n) | (x>>(-n&31));
}

static unsigned int
load32(const unsigned char* x)
{
  unsigned int u = x[3];
  u = (u<<8)|x[2];
  u = (u<<8)|x[1];
  return (u<<8)|x[0];
}

static void
store32(unsigned char *x,unsigned int u)
{
  int i;
  for (i=0; i < 4; ++i) { x[i] = u; u >>= 8; }
}

static void
qround(unsigned int* x,
       unsigned int a,
       unsigned int b,
       unsigned int c,
       unsigned int d)
{
  x[a] = x[a]+x[b]; x[d] = l32(x[d]^x[a],16);
  x[c] = x[c]+x[d]; x[b] = l32(x[b]^x[c],12);
  x[a] = x[a]+x[b]; x[d] = l32(x[d]^x[a], 8);
  x[c] = x[c]+x[d]; x[b] = l32(x[b]^x[c], 7);
}

static void
core(unsigned char out[64],
     const unsigned int in[16])
{
  unsigned int x[16];
  int i;

  /* Copy input block */
  for (i = 0; i < 16; ++i) { x[i] = in[i]; }

  /* Ten iterations, two rounds each iteration */
  for (i = 0; i < 10; ++i) {
    /* Column round */
    qround(x, 0, 4, 8,  12);
    qround(x, 1, 5, 9,  13);
    qround(x, 2, 6, 10, 14);
    qround(x, 3, 7, 11, 15);
    /* Diagonal round */
    qround(x, 0, 5, 10, 15);
    qround(x, 1, 6, 11, 12);
    qround(x, 2, 7, 8,  13);
    qround(x, 3, 4, 9,  14);
  }

  for (i = 0; i < 16; ++i) { x[i] = x[i] + in[i]; }      /* Add input */
  for (i = 0; i < 16; ++i) { store32(out+(4*i), x[i]); } /* Output */
}

static void
kexp(unsigned int* st,
     const unsigned int counter,
     const unsigned char* nonce,
     const unsigned char* key)
{
  /* Constants */
  st[0]  = 0x61707865;
  st[1]  = 0x3320646e;
  st[2]  = 0x79622d32;
  st[3]  = 0x6b206574;
  /* 4-11: key bytes */
  st[4]  = load32(key);
  st[5]  = load32(key+4);
  st[6]  = load32(key+8);
  st[7]  = load32(key+12);
  st[8]  = load32(key+16);
  st[9]  = load32(key+20);
  st[10] = load32(key+24);
  st[11] = load32(key+28);
  /* Counter + Nonce */
  st[12] = counter;
  st[13] = load32(nonce);
  st[14] = load32(nonce+4);
  st[15] = load32(nonce+8);
}

int
crypto_stream_chacha20_xor(unsigned char* out,
                           const unsigned char* msg,
                           unsigned long long mlen,
                           const unsigned int counter,
                           const unsigned char* nonce,
                           const unsigned char* key)
{
  unsigned long long i;
  unsigned int st[16];
  unsigned char blk[64];

  kexp(st, counter, nonce, key);

  for(;;) {
    /* Advance state */
    core(blk, st);
    st[12] = st[12]+1;

    /* Fast path */
    if (mlen <= 64) {
      for (i = 0; i < mlen; ++i) out[i] = msg[i] ^ blk[i];
      return 0;
    }

    /* Normal path */
    for (i = 0; i < 64; ++i) { out[i] = msg[i] ^ blk[i]; }
    mlen -= 64;
    msg += 64;
    out += 64;
  }
  return 0;
}

int
crypto_stream_chacha20(unsigned char* out,
                       unsigned long long olen,
                       unsigned int counter,
                       const unsigned char* nonce,
                       const unsigned char* key)
{
  /* Zero the output before encrypting it. */
  for (unsigned long long i=0; i < olen; ++i) { out[i] = 0; }

  return crypto_stream_chacha20_xor(out, out, olen, counter, nonce, key);
}

/* -------------------------------------------------------------------------- */
/* -- Testing machinery ----------------------------------------------------- */

#if defined(WITH_TESTS)
#include <stdio.h>
#include <stdlib.h>

/* Note: why don't we use memset/memcmp? CompCert interpreter compatibility. */
static void
clear(unsigned char* o, unsigned int n)
{
  unsigned int i = 0;
  for (; i < n; i++) o[i] = 0;
}

static int
verify(unsigned char* x, unsigned char* y, unsigned int n)
{
  unsigned int i = 0, d = 0;
  for(; i < n; i++) d |= x[i] ^ y[i];
  return (1 & ((d - 1) >> 8)) - 1;
}

/* -------------------------------------------------------------------------- */
/* -- Test vectors ---------------------------------------------------------- */

unsigned char msg1[64] = { 0, };
unsigned char out1[]
  = { 0x76, 0xb8, 0xe0, 0xad, 0xa0, 0xf1, 0x3d, 0x90, 0x40, 0x5d,
      0x6a, 0xe5, 0x53, 0x86, 0xbd, 0x28, 0xbd, 0xd2, 0x19, 0xb8,
      0xa0, 0x8d, 0xed, 0x1a, 0xa8, 0x36, 0xef, 0xcc, 0x8b, 0x77,
      0x0d, 0xc7, 0xda, 0x41, 0x59, 0x7c, 0x51, 0x57, 0x48, 0x8d,
      0x77, 0x24, 0xe0, 0x3f, 0xb8, 0xd8, 0x4a, 0x37, 0x6a, 0x43,
      0xb8, 0xf4, 0x15, 0x18, 0xa1, 0x1c, 0xc3, 0x87, 0xb6, 0x69,
      0xb2, 0xee, 0x65, 0x86 };

unsigned char msg2[]
  = { 0x41, 0x6e, 0x79, 0x20, 0x73, 0x75, 0x62, 0x6d, 0x69, 0x73,
      0x73, 0x69, 0x6f, 0x6e, 0x20, 0x74, 0x6f, 0x20, 0x74, 0x68,
      0x65, 0x20, 0x49, 0x45, 0x54, 0x46, 0x20, 0x69, 0x6e, 0x74,
      0x65, 0x6e, 0x64, 0x65, 0x64, 0x20, 0x62, 0x79, 0x20, 0x74,
      0x68, 0x65, 0x20, 0x43, 0x6f, 0x6e, 0x74, 0x72, 0x69, 0x62,
      0x75, 0x74, 0x6f, 0x72, 0x20, 0x66, 0x6f, 0x72, 0x20, 0x70,
      0x75, 0x62, 0x6c, 0x69, 0x63, 0x61, 0x74, 0x69, 0x6f, 0x6e,
      0x20, 0x61, 0x73, 0x20, 0x61, 0x6c, 0x6c, 0x20, 0x6f, 0x72,
      0x20, 0x70, 0x61, 0x72, 0x74, 0x20, 0x6f, 0x66, 0x20, 0x61,
      0x6e, 0x20, 0x49, 0x45, 0x54, 0x46, 0x20, 0x49, 0x6e, 0x74,
      0x65, 0x72, 0x6e, 0x65, 0x74, 0x2d, 0x44, 0x72, 0x61, 0x66,
      0x74, 0x20, 0x6f, 0x72, 0x20, 0x52, 0x46, 0x43, 0x20, 0x61,
      0x6e, 0x64, 0x20, 0x61, 0x6e, 0x79, 0x20, 0x73, 0x74, 0x61,
      0x74, 0x65, 0x6d, 0x65, 0x6e, 0x74, 0x20, 0x6d, 0x61, 0x64,
      0x65, 0x20, 0x77, 0x69, 0x74, 0x68, 0x69, 0x6e, 0x20, 0x74,
      0x68, 0x65, 0x20, 0x63, 0x6f, 0x6e, 0x74, 0x65, 0x78, 0x74,
      0x20, 0x6f, 0x66, 0x20, 0x61, 0x6e, 0x20, 0x49, 0x45, 0x54,
      0x46, 0x20, 0x61, 0x63, 0x74, 0x69, 0x76, 0x69, 0x74, 0x79,
      0x20, 0x69, 0x73, 0x20, 0x63, 0x6f, 0x6e, 0x73, 0x69, 0x64,
      0x65, 0x72, 0x65, 0x64, 0x20, 0x61, 0x6e, 0x20, 0x22, 0x49,
      0x45, 0x54, 0x46, 0x20, 0x43, 0x6f, 0x6e, 0x74, 0x72, 0x69,
      0x62, 0x75, 0x74, 0x69, 0x6f, 0x6e, 0x22, 0x2e, 0x20, 0x53,
      0x75, 0x63, 0x68, 0x20, 0x73, 0x74, 0x61, 0x74, 0x65, 0x6d,
      0x65, 0x6e, 0x74, 0x73, 0x20, 0x69, 0x6e, 0x63, 0x6c, 0x75,
      0x64, 0x65, 0x20, 0x6f, 0x72, 0x61, 0x6c, 0x20, 0x73, 0x74,
      0x61, 0x74, 0x65, 0x6d, 0x65, 0x6e, 0x74, 0x73, 0x20, 0x69,
      0x6e, 0x20, 0x49, 0x45, 0x54, 0x46, 0x20, 0x73, 0x65, 0x73,
      0x73, 0x69, 0x6f, 0x6e, 0x73, 0x2c, 0x20, 0x61, 0x73, 0x20,
      0x77, 0x65, 0x6c, 0x6c, 0x20, 0x61, 0x73, 0x20, 0x77, 0x72,
      0x69, 0x74, 0x74, 0x65, 0x6e, 0x20, 0x61, 0x6e, 0x64, 0x20,
      0x65, 0x6c, 0x65, 0x63, 0x74, 0x72, 0x6f, 0x6e, 0x69, 0x63,
      0x20, 0x63, 0x6f, 0x6d, 0x6d, 0x75, 0x6e, 0x69, 0x63, 0x61,
      0x74, 0x69, 0x6f, 0x6e, 0x73, 0x20, 0x6d, 0x61, 0x64, 0x65,
      0x20, 0x61, 0x74, 0x20, 0x61, 0x6e, 0x79, 0x20, 0x74, 0x69,
      0x6d, 0x65, 0x20, 0x6f, 0x72, 0x20, 0x70, 0x6c, 0x61, 0x63,
      0x65, 0x2c, 0x20, 0x77, 0x68, 0x69, 0x63, 0x68, 0x20, 0x61,
      0x72, 0x65, 0x20, 0x61, 0x64, 0x64, 0x72, 0x65, 0x73, 0x73,
      0x65, 0x64, 0x20, 0x74, 0x6f };
unsigned char out2[]
  = { 0xa3, 0xfb, 0xf0, 0x7d, 0xf3, 0xfa, 0x2f, 0xde, 0x4f, 0x37,
      0x6c, 0xa2, 0x3e, 0x82, 0x73, 0x70, 0x41, 0x60, 0x5d, 0x9f,
      0x4f, 0x4f, 0x57, 0xbd, 0x8c, 0xff, 0x2c, 0x1d, 0x4b, 0x79,
      0x55, 0xec, 0x2a, 0x97, 0x94, 0x8b, 0xd3, 0x72, 0x29, 0x15,
      0xc8, 0xf3, 0xd3, 0x37, 0xf7, 0xd3, 0x70, 0x05, 0x0e, 0x9e,
      0x96, 0xd6, 0x47, 0xb7, 0xc3, 0x9f, 0x56, 0xe0, 0x31, 0xca,
      0x5e, 0xb6, 0x25, 0x0d, 0x40, 0x42, 0xe0, 0x27, 0x85, 0xec,
      0xec, 0xfa, 0x4b, 0x4b, 0xb5, 0xe8, 0xea, 0xd0, 0x44, 0x0e,
      0x20, 0xb6, 0xe8, 0xdb, 0x09, 0xd8, 0x81, 0xa7, 0xc6, 0x13,
      0x2f, 0x42, 0x0e, 0x52, 0x79, 0x50, 0x42, 0xbd, 0xfa, 0x77,
      0x73, 0xd8, 0xa9, 0x05, 0x14, 0x47, 0xb3, 0x29, 0x1c, 0xe1,
      0x41, 0x1c, 0x68, 0x04, 0x65, 0x55, 0x2a, 0xa6, 0xc4, 0x05,
      0xb7, 0x76, 0x4d, 0x5e, 0x87, 0xbe, 0xa8, 0x5a, 0xd0, 0x0f,
      0x84, 0x49, 0xed, 0x8f, 0x72, 0xd0, 0xd6, 0x62, 0xab, 0x05,
      0x26, 0x91, 0xca, 0x66, 0x42, 0x4b, 0xc8, 0x6d, 0x2d, 0xf8,
      0x0e, 0xa4, 0x1f, 0x43, 0xab, 0xf9, 0x37, 0xd3, 0x25, 0x9d,
      0xc4, 0xb2, 0xd0, 0xdf, 0xb4, 0x8a, 0x6c, 0x91, 0x39, 0xdd,
      0xd7, 0xf7, 0x69, 0x66, 0xe9, 0x28, 0xe6, 0x35, 0x55, 0x3b,
      0xa7, 0x6c, 0x5c, 0x87, 0x9d, 0x7b, 0x35, 0xd4, 0x9e, 0xb2,
      0xe6, 0x2b, 0x08, 0x71, 0xcd, 0xac, 0x63, 0x89, 0x39, 0xe2,
      0x5e, 0x8a, 0x1e, 0x0e, 0xf9, 0xd5, 0x28, 0x0f, 0xa8, 0xca,
      0x32, 0x8b, 0x35, 0x1c, 0x3c, 0x76, 0x59, 0x89, 0xcb, 0xcf,
      0x3d, 0xaa, 0x8b, 0x6c, 0xcc, 0x3a, 0xaf, 0x9f, 0x39, 0x79,
      0xc9, 0x2b, 0x37, 0x20, 0xfc, 0x88, 0xdc, 0x95, 0xed, 0x84,
      0xa1, 0xbe, 0x05, 0x9c, 0x64, 0x99, 0xb9, 0xfd, 0xa2, 0x36,
      0xe7, 0xe8, 0x18, 0xb0, 0x4b, 0x0b, 0xc3, 0x9c, 0x1e, 0x87,
      0x6b, 0x19, 0x3b, 0xfe, 0x55, 0x69, 0x75, 0x3f, 0x88, 0x12,
      0x8c, 0xc0, 0x8a, 0xaa, 0x9b, 0x63, 0xd1, 0xa1, 0x6f, 0x80,
      0xef, 0x25, 0x54, 0xd7, 0x18, 0x9c, 0x41, 0x1f, 0x58, 0x69,
      0xca, 0x52, 0xc5, 0xb8, 0x3f, 0xa3, 0x6f, 0xf2, 0x16, 0xb9,
      0xc1, 0xd3, 0x00, 0x62, 0xbe, 0xbc, 0xfd, 0x2d, 0xc5, 0xbc,
      0xe0, 0x91, 0x19, 0x34, 0xfd, 0xa7, 0x9a, 0x86, 0xf6, 0xe6,
      0x98, 0xce, 0xd7, 0x59, 0xc3, 0xff, 0x9b, 0x64, 0x77, 0x33,
      0x8f, 0x3d, 0xa4, 0xf9, 0xcd, 0x85, 0x14, 0xea, 0x99, 0x82,
      0xcc, 0xaf, 0xb3, 0x41, 0xb2, 0x38, 0x4d, 0xd9, 0x02, 0xf3,
      0xd1, 0xab, 0x7a, 0xc6, 0x1d, 0xd2, 0x9c, 0x6f, 0x21, 0xba,
      0x5b, 0x86, 0x2f, 0x37, 0x30, 0xe3, 0x7c, 0xfd, 0xc4, 0xfd,
      0x80, 0x6c, 0x22, 0xf2, 0x21 };

unsigned char key3[]
  = { 0x1c, 0x92, 0x40, 0xa5, 0xeb, 0x55, 0xd3, 0x8a, 0xf3, 0x33,
      0x88, 0x86, 0x04, 0xf6, 0xb5, 0xf0, 0x47, 0x39, 0x17, 0xc1,
      0x40, 0x2b, 0x80, 0x09, 0x9d, 0xca, 0x5c, 0xbc, 0x20, 0x70,
      0x75, 0xc0 };
unsigned char msg3[]
  = { 0x27, 0x54, 0x77, 0x61, 0x73, 0x20, 0x62, 0x72, 0x69, 0x6c,
      0x6c, 0x69, 0x67, 0x2c, 0x20, 0x61, 0x6e, 0x64, 0x20, 0x74,
      0x68, 0x65, 0x20, 0x73, 0x6c, 0x69, 0x74, 0x68, 0x79, 0x20,
      0x74, 0x6f, 0x76, 0x65, 0x73, 0x0a, 0x44, 0x69, 0x64, 0x20,
      0x67, 0x79, 0x72, 0x65, 0x20, 0x61, 0x6e, 0x64, 0x20, 0x67,
      0x69, 0x6d, 0x62, 0x6c, 0x65, 0x20, 0x69, 0x6e, 0x20, 0x74,
      0x68, 0x65, 0x20, 0x77, 0x61, 0x62, 0x65, 0x3a, 0x0a, 0x41,
      0x6c, 0x6c, 0x20, 0x6d, 0x69, 0x6d, 0x73, 0x79, 0x20, 0x77,
      0x65, 0x72, 0x65, 0x20, 0x74, 0x68, 0x65, 0x20, 0x62, 0x6f,
      0x72, 0x6f, 0x67, 0x6f, 0x76, 0x65, 0x73, 0x2c, 0x0a, 0x41,
      0x6e, 0x64, 0x20, 0x74, 0x68, 0x65, 0x20, 0x6d, 0x6f, 0x6d,
      0x65, 0x20, 0x72, 0x61, 0x74, 0x68, 0x73, 0x20, 0x6f, 0x75,
      0x74, 0x67, 0x72, 0x61, 0x62, 0x65, 0x2e };
unsigned char out3[]
  = { 0x62, 0xe6, 0x34, 0x7f, 0x95, 0xed, 0x87, 0xa4, 0x5f, 0xfa,
      0xe7, 0x42, 0x6f, 0x27, 0xa1, 0xdf, 0x5f, 0xb6, 0x91, 0x10,
      0x04, 0x4c, 0x0d, 0x73, 0x11, 0x8e, 0xff, 0xa9, 0x5b, 0x01,
      0xe5, 0xcf, 0x16, 0x6d, 0x3d, 0xf2, 0xd7, 0x21, 0xca, 0xf9,
      0xb2, 0x1e, 0x5f, 0xb1, 0x4c, 0x61, 0x68, 0x71, 0xfd, 0x84,
      0xc5, 0x4f, 0x9d, 0x65, 0xb2, 0x83, 0x19, 0x6c, 0x7f, 0xe4,
      0xf6, 0x05, 0x53, 0xeb, 0xf3, 0x9c, 0x64, 0x02, 0xc4, 0x22,
      0x34, 0xe3, 0x2a, 0x35, 0x6b, 0x3e, 0x76, 0x43, 0x12, 0xa6,
      0x1a, 0x55, 0x32, 0x05, 0x57, 0x16, 0xea, 0xd6, 0x96, 0x25,
      0x68, 0xf8, 0x7d, 0x3f, 0x3f, 0x77, 0x04, 0xc6, 0xa8, 0xd1,
      0xbc, 0xd1, 0xbf, 0x4d, 0x50, 0xd6, 0x15, 0x4b, 0x6d, 0xa7,
      0x31, 0xb1, 0x87, 0xb5, 0x8d, 0xfd, 0x72, 0x8a, 0xfa, 0x36,
      0x75, 0x7a, 0x79, 0x7a, 0xc1, 0x88, 0xd1 };

/* -------------------------------------------------------------------------- */
/* -- Test driver ----------------------------------------------------------- */

int
main()
{
  int r = 0;
  unsigned char out[512];
  unsigned char key[32];
  unsigned char nonce[12];

#define WIPE()                 \
  clear(out,   sizeof(out));   \
  clear(key,   sizeof(key));   \
  clear(nonce, sizeof(nonce));

#define ENCRYPT(msg, cnt, nonce, key)                                   \
  crypto_stream_chacha20_xor(out, msg, sizeof(msg), cnt, nonce, key);

#define CHECK(buf, msg)                          \
  if (0 != verify(out, buf, sizeof(buf))) {      \
    fprintf(stderr, "FAILURE: %s", msg);         \
    exit(1);                                     \
  }

  /* ------------------------------------------------------------------------ */
  /* ---- Test vector #1 ---------------------------------------------------- */
  WIPE();

  ENCRYPT(msg1, 0, nonce, key);
  CHECK(out1, "Test vector #1 failed");

  /* ------------------------------------------------------------------------ */
  /* ---- Test vector #2 ---------------------------------------------------- */
  WIPE();
  key[31]   = 0x01;
  nonce[11] = 0x02;

  ENCRYPT(msg2, 1, nonce, key);
  CHECK(out2, "Test vector #2 failed");

  /* ------------------------------------------------------------------------ */
  /* ---- Test vector #3 ---------------------------------------------------- */
  WIPE();
  nonce[11] = 0x02;

  ENCRYPT(msg3, 42, nonce, key3);
  CHECK(out3, "Test vector #3 failed");

  /* ------------------------------------------------------------------------ */
  /* ---- Done -------------------------------------------------------------- */
  printf("OK\n");

#undef CHECK
#undef ENCRYPT
#undef WIPE
  return 0;
}

#endif /* defined(WITH_TESTS) */
